/***************************************************************************
 *   Copyright (C) 2007 by Subhrendu Sarkar   *
 *   ss3295@columbia.edu   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <sofia-sip/nua.h>
#include "main_window.hh"

typedef struct ssc_conf_s ssc_conf_t;


#include <sofia-sip/sip.h>
#include <sofia-sip/sip_header.h>
#include <sofia-sip/sip_status.h>
#include <sofia-sip/nua.h>
#include <sofia-sip/nua_tag.h>
#include <sofia-sip/soa.h>
#include <sofia-sip/su_tag_io.h>
#include <sofia-sip/su_tagarg.h>
#include <sofia-sip/sl_utils.h>
#include <sofia-sip/su_debug.h>

#include <sofia-sip/stun_tag.h>


#if HAVE_FUNC
#define enter (void)SU_DEBUG_9(("%s: entering\n", __func__))
#elif HAVE_FUNCTION
#define enter (void)SU_DEBUG_9(("%s: entering\n", __FUNCTION__))
#else
#define enter (void)0
#endif

/* define type of context pointers for callbacks */
typedef struct pua_s pua_t;
#define NUA_MAGIC_T pua_t
#define SOA_MAGIC_T pua_t


typedef struct ssc_oper_s ssc_oper_t;

/* define type of context pointers for callbacks */
#define NUA_IMAGIC_T    ssc_oper_t
#define NUA_HMAGIC_T    ssc_oper_t

typedef struct ssc_conf_s ssc_conf_t;

struct ssc_oper_s {
//  ssc_oper_t   *op_next;

  /**< Remote end identity
   *
   * Contents of To: when initiating, From: when receiving.
   */
  char const   *op_ident;	

  /** NUA handle */ 
  nua_handle_t *op_handle;

  /** How this handle was used initially */
  sip_method_t  op_method;	/**< REGISTER, INVITE, MESSAGE, or SUBSCRIBE */
  char const   *op_method_name;

  /** Call state. 
   *
   * - opc_sent when initial INVITE has been sent
   * - opc_recv when initial INVITE has been received
   * - opc_complate when 200 Ok has been sent/received
   * - opc_active when media is used
   * - opc_sent when re-INVITE has been sent
   * - opc_recv when re-INVITE has been received
   */
  enum { 
    opc_none, 
    opc_sent = 1, 
    opc_recv = 2, 
    opc_complete = 3, 
    opc_active = 4,
    opc_sent_hold = 8,             /**< Call put on hold */
    opc_pending = 16               /**< Waiting for local resources */
  } op_callstate;

  int           op_prev_state;     /**< Previous call state */

  unsigned      op_persistent : 1; /**< Is this handle persistent? */
  unsigned      op_referred : 1;
  unsigned :0;
};


/** 
 * Configuration data for pua_s_create().
 */
struct ssc_conf_s {
  const char   *ssc_aor;        /**< Public SIP address aka AOR (SIP URI) */
  const char   *ssc_certdir;	/**< Directory for TLS certs (directory path) */
  const char   *ssc_contact;	/**< SIP contact URI (local address to use) */
  const char   *ssc_media_addr;	/**< Media address (hostname, IP address) */
  const char   *ssc_media_impl;	/**< Media address (hostname, IP address) */
  const char   *ssc_proxy;	/**< SIP outbound proxy (SIP URI) */
  const char   *ssc_registrar;	/**< SIP registrar (SIP URI) */
  const char   *ssc_stun_server;/**< STUN server address (hostname, IP address) */
  //gboolean      ssc_autoanswer; /**< Whether to autoanswer to calls */
  //ssc_exit_cb   ssc_exit_cb;    /**< Callback to signal stack shutdown */
};
/**
 * PUA context information structure
 */
typedef struct pua_s {
  su_home_t    ssc_home[1];	/**< Our memory home */
  char const   *ssc_name;	/**< Our name */
  su_root_t    *ssc_root;       /**< Pointer to application root */
  nua_t        *ssc_nua;        /**< Pointer to NUA object */

  ssc_oper_t   *ssc_operations;	/**< Remote destinations */

  char         *ssc_address;    /**< Current AOR */
  ssc_conf_t    conf[1];  	/**< Config settings for ssc_sip.h */
} pua_t;

void pua_publish(pua_t *pua,int state, char *note);
void ssc_oper_assign(ssc_oper_t *op, sip_method_t method, char const *name);
void event_callback(nua_event_t event, int status, char const *phrase, nua_t *nua, nua_magic_t *magic, nua_handle_t *nh, nua_hmagic_t *hmagic, sip_t const *sip, tagi_t tags[]);

ssc_oper_t *ssc_oper_create(pua_t *ssc, 
			    sip_method_t method,
				char const *name,
			    char *address,
			    tag_type_t tag, tag_value_t value, ...);
extern int g_busy_state;
extern pthread_mutex_t BusyState_mutex;
extern pthread_cond_t BusyState_cond;
pua_t *global_pua;
ssc_oper_t *global_op;
int publish_flag = 0;
void free_thread2(void *msgp)
{
	su_root_break(global_pua->ssc_root) ;
	nua_shutdown(global_pua->ssc_nua);
	if(publish_flag == 1)
		nua_destroy(global_pua->ssc_nua);
	su_root_destroy(global_pua->ssc_root);
	su_deinit();
}

int sip_pua(void *param)
{
	main_window *win = (main_window *)param;
	pua_t pua[1] = {{{{sizeof(pua)}}}};
	global_pua = pua; 
	pua->ssc_name = win->m_ssc_name/*"ss3295"*/;
	//Initialize system utilities
	su_init();
	//initialize memory handling
	su_home_init(pua->ssc_home);
	//Initialize the root object
	pua->ssc_root = su_root_create(pua);
	std::cout <<"SIP- PRESENCE SERVER ADDRESS - " << win->m_sip_proxy<< std::endl;
	setenv("SOFSIP_ADDRESS", win->m_sip_addr, 1); /*Example "sip:vs2140@cs.columbia.edu"*/
	setenv("SOFSIP_PROXY",  win->m_sip_proxy, 1); /*Example "sip:128.59.19.216" "sip:paris.clic.cs.columbia.edu"*/	

	/* step: process environment variables */
	pua->conf->ssc_aor = getenv("SOFSIP_ADDRESS");
	pua->conf->ssc_proxy = getenv("SOFSIP_PROXY");
	pua->ssc_address = su_strdup(pua->ssc_home, pua->conf->ssc_aor);
	pua->conf->ssc_contact = "sip:ss3295@columbia.edu";

	 /* step: launch the SIP stack */
  	pua->ssc_nua = nua_create(pua->ssc_root, 
			    event_callback, pua,
			    TAG_NULL());
	

	if (pua->ssc_nua) 
	{
		nua_set_params(pua->ssc_nua,
			TAG_IF(pua->conf->ssc_proxy,
				NUTAG_PROXY(pua->conf->ssc_proxy)),
 		   NUTAG_ENABLEMESSAGE(1),
		   NUTAG_ENABLEINVITE(1),
		   NUTAG_AUTOALERT(1),
		   NUTAG_SESSION_TIMER(3600),
		  		TAG_IF(pua->conf->ssc_aor,
				SIPTAG_FROM_STR(pua->conf->ssc_aor)),
			TAG_NULL());
	}

	ssc_oper_t *op;
	sip_payload_t *pl = NULL;  
	char *address;
	char *xmlnote = NULL;
	int open;
	char *note = NULL;
	
	open = note == NULL || note[0] != '-';
	pthread_cleanup_push(free_thread2, NULL);

	while(1)
	{
		pthread_mutex_lock(&BusyState_mutex);
		pthread_cond_wait(&BusyState_cond, &BusyState_mutex);
		pua_publish(pua,g_busy_state, note);
		pthread_mutex_unlock(&BusyState_mutex);	
	
		su_root_run(pua->ssc_root);
	}
	int last_state,last_type,i = 0;
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &last_type);
	int s = pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &last_state);	

	pthread_cleanup_pop(0);
	nua_destroy(pua->ssc_nua);
	su_root_destroy(pua->ssc_root);
	su_deinit();
	return 0;	
}

void pua_publish(pua_t *pua,int state, char *note)
{
	int last_tmp;
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &last_tmp);
		
	ssc_oper_t *op;
	sip_payload_t *pl = NULL;
	char *address;
	char *xmlnote = NULL;
	int open;
	int state_busy = state;

	sleep(1);	
	pthread_testcancel(); 
	open = note == NULL || note[0] != '-';
	char *state_b;
	if(state_busy)
		state_b = "busy";
	else
		state_b = "idle";
	
	std::cout << "PUBLISH STATE : " << state_b << std::endl;
	if (note && strcmp(note, "-") != 0)
		xmlnote = su_sprintf(pua->ssc_home, "<note>%s</note>\n", 
			 open ? note : note + 1);

	pl = sip_payload_format
			(pua->ssc_home, 
			 "<?xml version='1.0' encoding='UTF-8'?>\n"
			 "<presence xmlns='urn:ietf:params:xml:ns:cpim-pidf'\n"
			 "          entity='%s'>\n"
			 "  <tuple id='%s'>\n"
			 "    <status><basic>%s</basic>\n</status>\n"
			 "%s"
			 "  </tuple>\n"
			 "</presence>\n",
			 pua->ssc_address, pua->ssc_name, 
			 state_busy ? "closed" : "open", 
			 xmlnote ? xmlnote : "");

  address = su_strdup(pua->ssc_home, pua->ssc_address);

  if ((op = ssc_oper_create(pua, SIP_METHOD_PUBLISH, address,			    SIPTAG_EVENT_STR("presence"),TAG_END())))
 {
	global_op = op;
    printf("%s: %s %s\n", pua->ssc_name, op->op_method_name, op->op_ident);
    nua_publish(op->op_handle, 
		SIPTAG_CONTENT_TYPE_STR("application/cpim-pidf+xml"),
		SIPTAG_PAYLOAD(pl),
		SIPTAG_EXPIRES_STR("3600"),
		TAG_END());
  }
  publish_flag = 1;
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &last_tmp);

}

ssc_oper_t *ssc_oper_create(pua_t *ssc, 
			    sip_method_t method,
			    char const *name,
			    char *address,
			    tag_type_t tag, tag_value_t value, ...)
{
  ssc_oper_t *op, *old;
  ta_list ta;
  enter;
  if (address) {
    int have_url = 1;
    sip_to_t *to;
    to = sip_to_make(ssc->ssc_home, (char *)address);
    if (to == NULL) {
      printf("%s: : invalid address: %s\n", ssc->ssc_name, address);
      return NULL;
    }

    /* Try to make sense out of the URL */
    if (url_sanitize(to->a_url) < 0) {
      printf("%s: %s: invalid address\n", ssc->ssc_name);
      return NULL;
    }

    if (!(op = (ssc_oper_t*)(su_zalloc(ssc->ssc_home, sizeof(*op)))) ) {
      printf("%s: %s: cannot create handle\n", ssc->ssc_name);
      return NULL;
    }

    ta_start(ta, tag, value); 
    op->op_handle = nua_handle(ssc->ssc_nua, op, 
			       TAG_IF(have_url, NUTAG_URL(to->a_url)), 
			       SIPTAG_TO(to),
			       ta_tags(ta));
    ta_end(ta);  
    op->op_ident = sip_header_as_string(ssc->ssc_home, (sip_header_t *)to);
	ssc_oper_assign(op, method, name);
    su_free(ssc->ssc_home, to);
  }
  else if (method || name) 
    ssc_oper_assign(op = old, method, name);
  else
    return old;

  if (!op) {
    if (address)
      printf("%s: %s: invalid destination\n", ssc->ssc_name);
    else
      printf("%s: %s: no destination\n", ssc->ssc_name);
    return NULL;
  }
  return op;
}

/**
 * Assigns flags to operation object based on method type
 */
void ssc_oper_assign(ssc_oper_t *op, sip_method_t method, char const *name)
{
  if (!op)
    return;

  op->op_method = method, op->op_method_name = name;

  op->op_persistent = 
    method == sip_method_subscribe ||
    method == sip_method_register ||
    method == sip_method_publish;
}


/**
 * Callback for an outgoing PUBLISH request.
 */
void ssc_r_publish(int status, char const *phrase, 
		   nua_t *nua, pua_t *ssc,
		   nua_handle_t *nh, ssc_oper_t *op, sip_t const *sip,
		   tagi_t tags[])
{
  printf("%s: PUBLISH: %03d %s\n", ssc->ssc_name, status, phrase);

  if (status < 200)
    return;

  if (status == 401 || status == 407)
	return;
//    ssc_store_pending_auth(ssc, op, sip, tags);
  else if (status >= 300)
  {
    if (op->op_handle)
   	 nua_handle_destroy(op->op_handle), op->op_handle = NULL;

  }	
  else if (!sip->sip_expires || sip->sip_expires->ex_delta == 0)
   {
	 if (op->op_handle)
    		nua_handle_destroy(op->op_handle), op->op_handle = NULL;

   }
}

void event_callback(nua_event_t event, int status, char const *phrase, nua_t *nua, nua_magic_t *pua, nua_handle_t *nh, nua_hmagic_t *op, sip_t const *sip, tagi_t tags[])
{

	tag_type_t tag = NULL;
  	tag_value_t value = 0;
	switch (event) {
  	case nua_r_get_params:
		printf("Event %s status %d %s\n", nua_event_name(event), status, phrase);
   	break;
	case nua_r_set_params:
		printf("Event %s status %d %s\n", nua_event_name(event), status, phrase);
   	break;
	case nua_r_publish:
		{
			ssc_r_publish(status, phrase, nua, (pua_t *)pua, nh, (ssc_oper_t *)op, sip, tags);
			pthread_mutex_lock(&BusyState_mutex);
			pthread_cond_wait(&BusyState_cond, &BusyState_mutex);
			pua_publish(global_pua,g_busy_state, NULL);
			pthread_mutex_unlock(&BusyState_mutex);	
		}
    	break;
	}
	
	
}

 



